import os
from collections import OrderedDict
from datetime import datetime, timedelta
import re
import logging
import pandas as pd

logger = logging.getLogger(__name__)


def dtv_parser(filepath: str, order='asc', as_datetime=True, remove_dups=True) -> dict:
    """Parse dtv type file

    :param filepath: Dtv filepath
    :param order: Timestamp order (asc or desc)
    :param as_datetime: return timestamps as datetime obj
    :return: {id: <id>, unit: <unit>, values: [(<timestamp>, <value>),..}
    """

    if order and not as_datetime:
        raise ValueError('Invalid arguments, `order` must be used together with `as_datetime`')

    with open(filepath, 'rb') as f:
        lines = []
        for line in f:
            decoded_line = line.decode(errors='ignore', encoding='windows-1252')
            lines.append(decoded_line)

    unit = (''.join(x.lstrip().rstrip() for x in lines[:6])).lstrip('Â')
    name, _ = os.path.splitext(os.path.basename(filepath))
    device = name.split('_')[0]
    channel = "_".join(name.split('_')[1:])
    device_id = f'{device}.{channel}'

    values = []

    for line in lines:
        try:
            match = re.match('(.*),(.*)', line.strip())
            if match:
                timestamp = match.group(1)
                if len(match.group(1).split('.')) == 1:
                    if as_datetime:
                        timestamp = datetime.strptime(match.group(1), '%d/%m/%Y %H:%M:%S')
                else:
                    if as_datetime:
                        timestamp = datetime.strptime(match.group(1), '%d/%m/%Y %H:%M:%S.%f')
                value = eval(match.group(2))
            else:
                continue
        except Exception as msg:
            logger.error('Parsing line:{}, e:'.format(line, msg))
            continue

        values.append((timestamp, value))

    if order and as_datetime:
        if order == 'asc':
            values = sorted(values, key=lambda x: x[0])
        if order == 'desc':
            values = sorted(values, key=lambda x: x[0], reverse=True)

    if remove_dups:
        values_dict = OrderedDict(values)
        values = list(values_dict.items())

    return {'id': device_id, 'unit': unit, 'values': values}


def dtv_write(filepath: str, values: list = None, unit: str = None) -> str:
    """ Write dtv file

    :param filepath: Dtv filepath
    :param values: Values to write
    :param unit: Unit
    :return: Dtv filepath
    """

    header = ['\n', '\n', unit, '\n', '\n', '\n', '\n', '\n', '\n', '\n']
    body = []
    for i in values:
        ts = i[0]
        val = i[1]
        ts_str = ts.strftime("%d/%m/%Y %H:%M:%S")
        body.append(f'{ts_str},{val}\n')
    with open(filepath, 'w') as f:
        f.writelines(header + body)
    return filepath


def dtv_to_pandas(filepath) -> pd.DataFrame:
    """ Convert dtv type file to pandas
    :param filepath:
    :return: Pandas dataframe
    """
    values = dtv_parser(filepath)
    if values['values']:
        d = {'timestamp': [i[0] for i in values['values']], 'values': [i[1] for i in values['values']]}
        df = pd.DataFrame(data=d).set_index(['timestamp'])
    else:
        df = pd.DataFrame(data={'timestamp': [], 'values': []})
    return df


def compensate_files(filepath: str, filepath_comp: str = None, filepath_out: str = None,
                     offset: float = None, operator='-', max_tolerance: float = None) -> str:
    """
    Compensate two files.
    Take file and find closest measurement in file_comp and perform math: file <operator> file_comp

    :param filepath: Filepath in
    :param filepath_comp: Filepath for compensation
    :param filepath_out: Filepath out
    :param offset: Data offset
    :param operator: Math operator between input file and compensation file (-,+,*)
    :param max_tolerance: Max timestamp tolerance [minutes]
    :return: result filepath
    """

    if not filepath:
        raise ValueError('Invalid value for `filepath`, must not be `None`')

    if not filepath_out:
        filename_out = f'{os.path.basename(filepath)}_kompenzirano.dtv'
        filepath_out = os.path.join(os.path.dirname(filepath), filename_out)

    data_in = dtv_parser(filepath=filepath, as_datetime=True)

    df_comp = None
    if filepath_comp:
        df_comp = dtv_to_pandas(filepath=filepath_comp)
        if df_comp.empty:
            raise ValueError(f'File `{filepath_comp}` is empty')

    values_out = []
    for i in data_in['values']:
        ts = i[0]
        val = i[1]
        val_out = val
        if filepath_comp:
            try:
                if max_tolerance is not None:
                    comp_index = df_comp.index.get_loc(ts, method='nearest', tolerance=timedelta(minutes=max_tolerance))
                else:
                    comp_index = df_comp.index.get_loc(ts, method='nearest')
                val_comp = df_comp.values[comp_index][0]
                val_out = eval(f'{val} {operator} {val_comp}')
            except KeyError:
                # No match was found within range
                logger.warning(f'No match was found in file `{filepath_comp}` within tolerance {max_tolerance}s')
                continue

        if offset:
            val_out += offset

        val_out = round(val_out, 3)

        values_out.append((ts, val_out))

    return dtv_write(filepath=filepath_out, unit=data_in['unit'], values=values_out)


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('-filepath', '-in', default=None, help="Input file")
    parser.add_argument('-filepath_comp', '-comp', default=None, help="Compensation file")
    parser.add_argument('-filepath_out', '-out', default=None, help="Output file")
    parser.add_argument('-offset', default=None, help="Offset")
    parser.add_argument('-max_tolerance', default=None, help="Tolerance")
    args = parser.parse_args()
    compensate_files(**vars(args))



from serial import SerialException
from serial.tools import list_ports
import serial
import logging
from sys import platform

logger = logging.getLogger(__name__)

def comports():
    """Return list of available comports"""

    available_ports = []
    ports = list_ports.comports()
    for port in ports:
        available_ports.append(str(port))
    return available_ports


def serial_ports():
    """Filter serial ports from all comports and strip away description of port"""

    com_ports = comports()
    # ser_ports = list(filter(lambda x: 'Serial Port' in x, ports))
    serial_ports = []
    for port in com_ports:
        if platform.startswith('win'):
            if 'Serial Port' in port:
                serial_ports.append(port.split(' - ')[0])
        if platform == 'darwin':
            if not 'Bluetooth' in port:
                serial_ports.append(port.split(' - ')[0])
        if platform == 'linux':
            serial_ports.append(port.split(' - ')[0])
    return serial_ports


def autodetermine_port():
    """Auto select first serial port"""

    ser_ports = serial_ports()
    if ser_ports:
        return serial_ports()[0]


def open_port(func):
    def wrapper(self, *args, **kwargs):
        if self.ser is None:
            self.open()
        return func(self, *args, **kwargs)
    return wrapper


class SerialComm(object):
    def __init__(self):
        self.port = None
        self.baudrate = 38400
        self.timeout = 0.1
        self.n_retry = 3
        self.ser = None

    def set(self, port: str = None, baudrate: int = 38400, timeout: float = 0.5):
        self.port = port if port else autodetermine_port()
        self.baudrate = baudrate
        self.timeout = timeout
        self.open()

    def open(self):
        """Connect to serial port"""

        if self.port is None:
            raise Exception('Serial port is not defined')

        if self.ser:
            self.close_port()

        try:
            self.ser = serial.Serial(
                port=self.port,
                baudrate=self.baudrate,
                parity=serial.PARITY_NONE,
                stopbits=serial.STOPBITS_ONE,
                bytesize=serial.EIGHTBITS,
                timeout=self.timeout
            )
            self.ser.flushInput()
            logger.debug(f"Serial port opened:{self.port}, baud:{self.baudrate}")
            logger.debug(f"Serial port flushed")
        except SerialException as e:
            logger.error(f"Failed to open serial port:{self.port}, error:{e}")
            if 'PermissionError' in str(e):
                raise SerialException('Serial port is already in use')
            if 'FileNotFoundError' in str(e):
                raise SerialException('Serial port is not available')
            raise

    def close_port(self):
        """Close serial port"""

        self.ser.close()
        self.ser = None

    @open_port
    def write(self, data: bytes = None):
        """Write bytes to serial port"""

        if not data:
            raise ValueError('No bytes to send')

        logger.debug(f"Data sent: {data}")
        if data:
            try:
                self.ser.write(data)
            except SerialException as e:
                logger.warning(e)
                self.close_port()
                raise

    @open_port
    def read(self, last_char=b'\x83'):
        """Read from serial 1 byte at a time. Stop when last_char is received."""

        frame = b""
        while True:
            try:
                byte = self.ser.read(1)
            except serial.SerialException as e:
                logger.warning(e)
                self.close_port()
                raise
            if byte:
                frame += byte
                if byte == last_char:
                    break
            else:
                break
        logger.debug(f"Data received: {frame}")
        return frame

    def readline(self) -> bytes:
        """
        Read line from serial port
        :return: Bytes
        """

        while True:
            return self.ser.readline()



import collections
import sys

file1 = sys.argv[1]
file2 = sys.argv[2]

content1 = open(file1, 'r', encoding='cp855').read()
content2 = open(file2, 'r', encoding='cp855').read()

cnt1 = collections.Counter(content1)
cnt2 = collections.Counter(content2)

for key, val in cnt1.items():
    if val != cnt2[key]:
        print (f'Difference key:{repr(key)} file1:{val}, file2:{cnt2[key]}')



from binascii import unhexlify

from .const import (
    WAKE_UP_DEVICE,
    START_MARKER_BYTES,
    END_MARKER_BYTES,
    SERVER_ADDR_BYTES,
    VALUE_WRITE
)
from .crc import calculate_crc16
from .exceptions import (
    NoAnswer,
    InvalidAnswer,
    CRCError,
    InvalidCommand,
    DeviceTimeout,
)


def replace_repeating_chars(data: bytes = None, reverse=False) -> bytes:
    """Replace possible repeating chars within communication package
    [0x82] -> [0x84][‘2’]
    [0x83] -> [0x84][‘3’]
    [0x84] -> [0x84][‘4’]

    if reverse:
    [0x82][‘2’] -> [0x82]
    [0x83][‘2’] -> [0x83]
    [0x84][‘2’] -> [0x84]

    :param data: Input communication package (bytes)
    :param reverse: Do a reverse replacing
    :return: Bytes of data
    """
    if reverse:
        data_out = b''
        skip_iter = False
        for i, ch in enumerate(data):
            if skip_iter:
                skip_iter = False
                continue
            if hex(ch) == '0x84':
                if len(data) > (i + 1):
                    skip_iter = True
                    if hex(data[i + 1]) == '0x34':
                        data_out += b'\x84'
                    elif hex(data[i + 1]) == '0x33':
                        data_out += b'\x83'
                    elif hex(data[i + 1]) == '0x32':
                        data_out += b'\x82'
                    else:
                        skip_iter = False
            else:
                data_out += bytes([ch])
    else:
        data_out = data
        data_out = data_out.replace(b'\x84', b'\x844')
        data_out = data_out.replace(b'\x83', b'\x843')
        data_out = data_out.replace(b'\x82', b'\x842')
    return data_out

def data_decompress(data:bytes) -> bytes:
    """Decompress data. Repeating set of chars are compressed into shorter set.
    Set of K equal chars is transformed (K=char, X=number of repeats):
    [F0][X]K ; K=4-239
    [F0][3][K-hi][K-low]X ; K=240-65520

    :data: Bytes to decompress
    :return: Decompress data bytes
     """

    data_out = b''
    i = 0
    while i < len(data):
        ch = data[i]
        if hex(ch) == '0xf0':
            if len(data) > (i + 1):
                n_repeats = data[i + 1]
                if n_repeats == 0:
                    data_out += bytes([ch])
                    i += 2
                    continue
                char = data[i + 2]
                data_out += n_repeats * bytes([char])
                i += 3
                continue
        data_out += bytes([ch])
        i += 1

    return data_out


def define_package(recv_addr: int, app_cmd: str, var_name: str = None, var_val: str = None, counter: int = 0,
                   rf: bool = False) -> bytes:
    """
    Returns string that represent whole communication package. Output string can be directly sent to device.
    Example: read real values from ism (addr 4) "820004FF80000000008101565252563D89C483"

    :param recv_addr: receiver address (e.g. 1000)
    :param app_cmd: application command (e.g. VR - value read)
    :param var_name: variable name (e.g. RV - real values)
    :param var_val: variable value (e.g. `Merilno mest` or `3600`)
    :param rf: RF or not (wifi or wired connection)
    :param counter: package counter (0-255)
    :return: parsed data as string
    """
    if not recv_addr:
        raise ValueError("Receiver address can't be none")

    # Generate receiver address from integer recv_addr
    if rf:
        recv_addr_complete = '{:8X}'.format(recv_addr).replace(' ', '0')
    else:
        recv_addr_1 = '{:2X}'.format(0).replace(' ', '0')
        recv_addr_2 = '{:2X}'.format(recv_addr).replace(' ', '0')
        recv_addr_3 = '{:2X}'.format(255).replace(' ', '0')
        recv_addr_4 = '{:2X}'.format(128).replace(' ', '0')
        recv_addr_complete = recv_addr_1 + recv_addr_2 + recv_addr_3 + recv_addr_4

    # Consts
    recv_addr_complete = unhexlify(recv_addr_complete)
    data_layer = app_cmd
    if var_name:
        data_layer += var_name
    if app_cmd == VALUE_WRITE:
        data_layer += '='
    if var_val:
        data_layer += var_val
    data_layer = data_layer.encode('iso-8859-2')
    if rf and WAKE_UP_DEVICE in app_cmd:
        data_layer += b'\x00'

    # Get hex counter value and 16bit CRC code
    counter_hex = unhexlify('81') + chr(counter).encode('iso-8859-2')
    data_for_crc = recv_addr_complete + SERVER_ADDR_BYTES + counter_hex + data_layer
    crc16 = unhexlify(calculate_crc16(data_for_crc))

    # Combine all data layers
    package = recv_addr_complete + SERVER_ADDR_BYTES + \
              counter_hex + data_layer + crc16
    package = replace_repeating_chars(data=package)
    return START_MARKER_BYTES + package + END_MARKER_BYTES


def data_to_parsed_string(data: bytes):
    """Parse data to list. Separate by comma.
    Example of a raw data: b'\x82\x00\x00\x00\x00\x00\x00\x00\x00\x01@O0.638,0.00,21.67,\to\x83
    """
    data_str = data.decode('iso-8859-2')
    if 'Name' and 'Com' in data_str:
        data_str = list(filter(None, data_str.split(',')))
        data_out = []
        for i, item in enumerate(data_str):
            # If first letter after comma is not in the upper case,
            # merge item with previous item
            if item[0].isupper():
                data_out.append(item)
            else:
                data_out[-1] += ',' + item
        data_str = data_out
    else:
        data_str = list(filter(None, data_str.split(',')))

    return data_str[0] if len(data_str) == 1 else data_str


def validate_data(data: bytes, stream_channel: bool = False) -> dict:
    '''Validate data bytes.
    :param data: input bytes
    :param stream_channel: stream channel uses different status_sign (O, E, OT or OM)
    '''

    if not data:
        raise NoAnswer('No answer from device')
    if len(data) < 14:
        raise InvalidAnswer('Data length is not sufficient')

    start_char = data[0]
    end_char = data[-1]
    data_layer = data[1:-3]
    crc = data[-3:-1]
    crc_ref = unhexlify(calculate_crc16(data_layer))
    if crc != crc_ref or start_char != 130 or end_char != 131:
        raise InvalidAnswer('Invalid answer')

    # Validate data. Possible combinations are:
    # sign @ signals start of message. First letter or first two letters after @ = status:
    # O[value] - okay
    # E[error message] - error
    # OM[data] -> okay, more data
    # OT[data] -> okay, data read complete
    # OE empty dir
    # W[percent] - not yet completed

    at_sign_pos = data_layer.find(b'@')
    if at_sign_pos == -1 or len(data_layer) == at_sign_pos:
        # Error in message if @ is not present or @ is last char in message
        raise InvalidAnswer('Unkown error:', data_layer)

    status_sign_pos = at_sign_pos + 1
    status_sign = chr(data_layer[status_sign_pos])
    if status_sign == 'O' or status_sign == 'W':
        if stream_channel and len(data_layer[status_sign_pos:]) > 1:
            status_sign_additional_letter = chr(data_layer[status_sign_pos + 1])
            if status_sign_additional_letter == 'T' or status_sign_additional_letter == 'M':
                status_sign += status_sign_additional_letter
        data_out = {'status': status_sign,
                    'data': data_layer[data_layer.find(status_sign.encode()) + len(status_sign):]}
        return data_out

    if status_sign == 'E':
        error_code = chr(data_layer[status_sign_pos + 1])
        error_msg = data_layer[status_sign_pos + 1:]
        if b'implemented' in data_layer:
            raise InvalidCommand('Invalid command')
        elif b'Undefined variable' in data_layer:
            raise InvalidCommand('Undefined variable')
        elif error_code == 'r':
            raise CRCError('Invalid CRC code')
        elif error_code == 'R':
            raise DeviceTimeout
        else:
            raise InvalidAnswer('Unkown error:', error_msg)
    elif status_sign == ' ':
        if b'implemented' in data_layer:
            raise InvalidCommand('Invalid command')
    else:
        raise InvalidAnswer('Status sign is not valid: ', data_layer)
